@@ -27,8 +27,7 @@
  * @see_also: #GstPlugin, #GstPluginFeature
  *
  * One registry holds the metadata of a set of plugins.
- * All registries build the #GstRegistryPool.
- * 
+ *
  * <emphasis role="bold">Design:</emphasis>
  *
  * The #GstRegistry object is a list of plugins and some functions for dealing
@@ -101,6 +100,8 @@
 #include <stdio.h>
 #include <string.h>
 
+/* For g_stat () */
+#include <glib/gstdio.h>
 
 #include "gstinfo.h"
 #include "gstregistry.h"
@@ -109,7 +110,7 @@
 
 #define GST_CAT_DEFAULT GST_CAT_REGISTRY
 
-/* the one instance of the default registry and the mutex protecting the 
+/* the one instance of the default registry and the mutex protecting the
  * variable. */
 static GStaticMutex _gst_registry_mutex = G_STATIC_MUTEX_INIT;
 static GstRegistry *_gst_registry_default = NULL;
@@ -122,8 +123,6 @@ enum
   LAST_SIGNAL
 };
 
-static void gst_registry_class_init (GstRegistryClass * klass);
-static void gst_registry_init (GstRegistry * registry);
 static void gst_registry_finalize (GObject * object);
 
 static guint gst_registry_signals[LAST_SIGNAL] = { 0 };
@@ -145,10 +144,27 @@ gst_registry_class_init (GstRegistryClass * klass)
 
   parent_class = g_type_class_peek_parent (klass);
 
+  /**
+   * GstRegistry::plugin-added:
+   * @registry: the registry that emitted the signal
+   * @plugin: the plugin that has been added
+   *
+   * Signals that a plugin has been added to the registry (possibly
+   * replacing a previously-added one by the same name)
+   */
   gst_registry_signals[PLUGIN_ADDED] =
       g_signal_new ("plugin-added", G_TYPE_FROM_CLASS (klass),
       G_SIGNAL_RUN_LAST, G_STRUCT_OFFSET (GstRegistryClass, plugin_added), NULL,
       NULL, gst_marshal_VOID__POINTER, G_TYPE_NONE, 1, G_TYPE_POINTER);
+
+  /**
+   * GstRegistry::feature-added:
+   * @registry: the registry that emitted the signal
+   * @feature: the feature that has been added
+   *
+   * Signals that a feature has been added to the registry (possibly
+   * replacing a previously-added one by the same name)
+   */
   gst_registry_signals[FEATURE_ADDED] =
       g_signal_new ("feature-added", G_TYPE_FROM_CLASS (klass),
       G_SIGNAL_RUN_LAST, G_STRUCT_OFFSET (GstRegistryClass, feature_added),
@@ -160,6 +176,7 @@ gst_registry_class_init (GstRegistryClass * klass)
 static void
 gst_registry_init (GstRegistry * registry)
 {
+  registry->feature_hash = g_hash_table_new (g_str_hash, g_str_equal);
 }
 
 static void
@@ -178,9 +195,9 @@ gst_registry_finalize (GObject * object)
     GstPlugin *plugin = p->data;
 
     if (plugin) {
-      GST_DEBUG_OBJECT (registry, "removing plugin %s",
+      GST_LOG_OBJECT (registry, "removing plugin %s",
           gst_plugin_get_name (plugin));
-      gst_registry_remove_plugin (registry, plugin);
+      gst_object_unref (plugin);
     }
     p = g_list_next (p);
   }
@@ -194,12 +211,16 @@ gst_registry_finalize (GObject * object)
     GstPluginFeature *feature = f->data;
 
     if (feature) {
-      gst_registry_remove_feature (registry, feature);
+      GST_LOG_OBJECT (registry, "removing feature %p (%s)",
+          feature, gst_plugin_feature_get_name (feature));
+      gst_object_unref (feature);
     }
     f = g_list_next (f);
   }
   g_list_free (features);
 
+  g_hash_table_destroy (registry->feature_hash);
+  registry->feature_hash = NULL;
 
   G_OBJECT_CLASS (parent_class)->finalize (object);
 }
@@ -336,7 +357,7 @@ gst_registry_add_plugin (GstRegistry * registry, GstPlugin * plugin)
   gst_object_sink (plugin);
   GST_OBJECT_UNLOCK (registry);
 
-  GST_DEBUG_OBJECT (registry, "emitting plugin-added for filename \"%s\"",
+  GST_LOG_OBJECT (registry, "emitting plugin-added for filename \"%s\"",
       GST_STR_NULL (plugin->filename));
   g_signal_emit (G_OBJECT (registry), gst_registry_signals[PLUGIN_ADDED], 0,
       plugin);
@@ -344,6 +365,34 @@ gst_registry_add_plugin (GstRegistry * registry, GstPlugin * plugin)
   return TRUE;
 }
 
+static void
+gst_registry_remove_features_for_plugin_unlocked (GstRegistry * registry,
+    GstPlugin * plugin)
+{
+  GList *f;
+
+  g_return_if_fail (GST_IS_REGISTRY (registry));
+  g_return_if_fail (GST_IS_PLUGIN (plugin));
+
+  /* Remove all features for this plugin */
+  f = registry->features;
+  while (f != NULL) {
+    GList *next = g_list_next (f);
+    GstPluginFeature *feature = f->data;
+
+    if (feature && !strcmp (feature->plugin_name, gst_plugin_get_name (plugin))) {
+      GST_DEBUG_OBJECT (registry, "removing feature %p (%s) for plugin %s",
+          feature, gst_plugin_feature_get_name (feature),
+          gst_plugin_get_name (plugin));
+
+      registry->features = g_list_delete_link (registry->features, f);
+      g_hash_table_remove (registry->feature_hash, feature->name);
+      gst_object_unref (feature);
+    }
+    f = next;
+  }
+}
+
 /**
  * gst_registry_remove_plugin:
  * @registry: the registry to remove the plugin from
@@ -359,8 +408,12 @@ gst_registry_remove_plugin (GstRegistry * registry, GstPlugin * plugin)
   g_return_if_fail (GST_IS_REGISTRY (registry));
   g_return_if_fail (GST_IS_PLUGIN (plugin));
 
+  GST_DEBUG_OBJECT (registry, "removing plugin %p (%s)",
+      plugin, gst_plugin_get_name (plugin));
+
   GST_OBJECT_LOCK (registry);
   registry->plugins = g_list_remove (registry->plugins, plugin);
+  gst_registry_remove_features_for_plugin_unlocked (registry, plugin);
   GST_OBJECT_UNLOCK (registry);
   gst_object_unref (plugin);
 }
@@ -391,21 +444,28 @@ gst_registry_add_feature (GstRegistry * registry, GstPluginFeature * feature)
   existing_feature = gst_registry_lookup_feature_locked (registry,
       feature->name);
   if (G_UNLIKELY (existing_feature)) {
-    GST_DEBUG_OBJECT (registry, "Replacing existing feature %p (%s)",
+    GST_DEBUG_OBJECT (registry, "replacing existing feature %p (%s)",
         existing_feature, feature->name);
+    /* Remove the existing feature from the list now, before we insert the new
+     * one, but don't unref yet because the hash is still storing a reference to     * it. */
     registry->features = g_list_remove (registry->features, existing_feature);
-    gst_object_unref (existing_feature);
   }
 
   GST_DEBUG_OBJECT (registry, "adding feature %p (%s)", feature, feature->name);
 
   registry->features = g_list_prepend (registry->features, feature);
+  g_hash_table_replace (registry->feature_hash, feature->name, feature);
+
+  if (G_UNLIKELY (existing_feature)) {
+    /* We unref now. No need to remove the feature name from the hash table, it      * got replaced by the new feature */
+    gst_object_unref (existing_feature);
+  }
 
   gst_object_ref (feature);
   gst_object_sink (feature);
   GST_OBJECT_UNLOCK (registry);
 
-  GST_DEBUG_OBJECT (registry, "emitting feature-added for %s", feature->name);
+  GST_LOG_OBJECT (registry, "emitting feature-added for %s", feature->name);
   g_signal_emit (G_OBJECT (registry), gst_registry_signals[FEATURE_ADDED], 0,
       feature);
 
@@ -432,6 +492,7 @@ gst_registry_remove_feature (GstRegistry * registry, GstPluginFeature * feature)
 
   GST_OBJECT_LOCK (registry);
   registry->features = g_list_remove (registry->features, feature);
+  g_hash_table_remove (registry->feature_hash, feature->name);
   GST_OBJECT_UNLOCK (registry);
   gst_object_unref (feature);
 }
@@ -485,7 +546,8 @@ gst_registry_plugin_filter (GstRegistry * registry,
  * If the first flag is set, only the first match is
  * returned (as a list with a single object).
  *
- * Returns: a GList of plugin features, gst_plugin_feature_list_free after use.
+ * Returns: a #GList of #GstPluginFeature. Use gst_plugin_feature_list_free()
+ * after usage.
  *
  * MT safe.
  */
@@ -592,8 +654,8 @@ gst_registry_find_feature (GstRegistry * registry, const gchar * name,
  *
  * Retrieves a #GList of #GstPluginFeature of @type.
  *
- * Returns: a #GList of #GstPluginFeature of @type. gst_plugin_feature_list_free
- * after usage.
+ * Returns: a #GList of #GstPluginFeature of @type. Use
+ * gst_plugin_feature_list_free() after usage.
  *
  * MT safe.
  */
@@ -620,7 +682,7 @@ gst_registry_get_feature_list (GstRegistry * registry, GType type)
  * Get a copy of all plugins registered in the given registry. The refcount
  * of each element in the list in incremented.
  *
- * Returns: a #GList of #GstPlugin. gst_plugin_list_free after use.
+ * Returns: a #GList of #GstPlugin. Use gst_plugin_list_free() after usage.
  *
  * MT safe.
  */
@@ -645,20 +707,10 @@ gst_registry_get_plugin_list (GstRegistry * registry)
 static GstPluginFeature *
 gst_registry_lookup_feature_locked (GstRegistry * registry, const char *name)
 {
-  GList *g;
-  GstPluginFeature *feature;
-
   if (G_UNLIKELY (name == NULL))
     return NULL;
 
-  for (g = registry->features; g; g = g_list_next (g)) {
-    feature = GST_PLUGIN_FEATURE_CAST (g->data);
-    if (feature->name && strcmp (name, feature->name) == 0) {
-      return feature;
-    }
-  }
-
-  return NULL;
+  return g_hash_table_lookup (registry->feature_hash, name);
 }
 
 /**
@@ -667,7 +719,7 @@ gst_registry_lookup_feature_locked (GstRegistry * registry, const char *name)
  * @name: a #GstPluginFeature name
  *
  * Find a #GstPluginFeature with @name in @registry.
- * 
+ *
  * Returns: a #GstPluginFeature with its refcount incremented, use
  * gst_object_unref() after usage.
  *
@@ -701,6 +753,7 @@ gst_registry_lookup_locked (GstRegistry * registry, const char *filename)
     return NULL;
 
   basename = g_path_get_basename (filename);
+  /* FIXME: use GTree speed up lookups */
   for (g = registry->plugins; g; g = g_list_next (g)) {
     plugin = GST_PLUGIN_CAST (g->data);
     if (plugin->basename && strcmp (basename, plugin->basename) == 0) {
@@ -757,46 +810,61 @@ gst_registry_scan_path_level (GstRegistry * registry, const gchar * path,
     return FALSE;
 
   while ((dirent = g_dir_read_name (dir))) {
+    struct stat file_status;
+
     filename = g_strjoin ("/", path, dirent, NULL);
+    if (g_stat (filename, &file_status) < 0) {
+      /* Plugin will be removed from cache after the scan completes if it
+       * is still marked 'cached' */
+      g_free (filename);
+      continue;
+    }
 
-    GST_LOG_OBJECT (registry, "examining file: %s", filename);
-
-    if (g_file_test (filename, G_FILE_TEST_IS_DIR)) {
+    if (file_status.st_mode & S_IFDIR) {
+      /* skip the .debug directory, these contain elf files that are not
+       * useful or worse, can crash dlopen () */
+      if (g_str_equal (dirent, ".debug")) {
+        GST_LOG_OBJECT (registry, "found .debug directory, ignoring");
+        g_free (filename);
+        continue;
+      }
+      /* FIXME 0.11: Don't recurse into directories, this behaviour
+       * is inconsistent with other PATH environment variables
+       */
       if (level > 0) {
-        GST_LOG_OBJECT (registry, "found directory, recursing");
+        GST_LOG_OBJECT (registry, "recursing into directory %s", filename);
         changed |= gst_registry_scan_path_level (registry, filename, level - 1);
       } else {
-        GST_LOG_OBJECT (registry,
-            "found directory, but recursion level is too deep");
+        GST_LOG_OBJECT (registry, "not recursing into directory %s, "
+            "recursion level too deep", filename);
       }
       g_free (filename);
       continue;
     }
-    if (!g_file_test (filename, G_FILE_TEST_IS_REGULAR)) {
-      GST_LOG_OBJECT (registry, "not a regular file, ignoring");
+    if (!(file_status.st_mode & S_IFREG)) {
+      GST_LOG_OBJECT (registry, "%s is not a regular file, ignoring", filename);
       g_free (filename);
       continue;
     }
-    if (!g_str_has_suffix (filename, ".so") &&
-        !g_str_has_suffix (filename, ".dll") &&
-        !g_str_has_suffix (filename, ".dynlib")) {
-      GST_LOG_OBJECT (registry,
-          "extension is not recognized as module file, ignoring");
+    if (!g_str_has_suffix (dirent, G_MODULE_SUFFIX)
+#ifdef GST_EXTRA_MODULE_SUFFIX
+        && !g_str_has_suffix (dirent, GST_EXTRA_MODULE_SUFFIX)
+#endif
+        ) {
+      GST_LOG_OBJECT (registry, "extension is not recognized as module file, "
+          "ignoring file %s", filename);
       g_free (filename);
       continue;
     }
 
+    GST_LOG_OBJECT (registry, "file %s looks like a possible module", filename);
+
     /* plug-ins are considered unique by basename; if the given name
      * was already seen by the registry, we ignore it */
     plugin = gst_registry_lookup (registry, filename);
     if (plugin) {
-      struct stat file_status;
+      gboolean env_vars_changed, deps_changed = FALSE;
 
-      if (stat (filename, &file_status)) {
-        /* FIXME remove from cache */
-        g_free (filename);
-        continue;
-      }
       if (plugin->registered) {
         GST_DEBUG_OBJECT (registry,
             "plugin already registered from path \"%s\"",
@@ -805,19 +873,36 @@ gst_registry_scan_path_level (GstRegistry * registry, const gchar * path,
         gst_object_unref (plugin);
         continue;
       }
+
+      env_vars_changed = _priv_plugin_deps_env_vars_changed (plugin);
+
       if (plugin->file_mtime == file_status.st_mtime &&
-          plugin->file_size == file_status.st_size) {
-        GST_DEBUG_OBJECT (registry, "file %s cached", filename);
+          plugin->file_size == file_status.st_size && !env_vars_changed &&
+          !(deps_changed = _priv_plugin_deps_files_changed (plugin))) {
+        GST_LOG_OBJECT (registry, "file %s cached", filename);
         plugin->flags &= ~GST_PLUGIN_FLAG_CACHED;
-        GST_DEBUG_OBJECT (registry, "marking plugin %p as registered", plugin);
+        GST_LOG_OBJECT (registry, "marking plugin %p as registered as %s",
+            plugin, filename);
         plugin->registered = TRUE;
+        /* Update the file path on which we've seen this cached plugin
+         * to ensure the registry cache will reflect up to date information */
+        if (strcmp (plugin->filename, filename) != 0) {
+          g_free (plugin->filename);
+          plugin->filename = g_strdup (filename);
+          changed = TRUE;
+        }
       } else {
         GST_INFO_OBJECT (registry, "cached info for %s is stale", filename);
         GST_DEBUG_OBJECT (registry, "mtime %ld != %ld or size %"
-            G_GSIZE_FORMAT " != %"
-            G_GSIZE_FORMAT, plugin->file_mtime, file_status.st_mtime,
-            plugin->file_size, file_status.st_size);
+            G_GINT64_FORMAT " != %" G_GINT64_FORMAT " or external dependency "
+            "env_vars changed: %d or external dependencies changed: %d",
+            plugin->file_mtime, file_status.st_mtime,
+            (gint64) plugin->file_size, (gint64) file_status.st_size,
+            env_vars_changed, deps_changed);
         gst_registry_remove_plugin (gst_registry_get_default (), plugin);
+        /* We don't use a GError here because a failure to load some shared 
+         * objects as plugins is normal (particularly in the uninstalled case)
+         */
         newplugin = gst_plugin_load_file (filename, NULL);
         if (newplugin) {
           GST_DEBUG_OBJECT (registry, "marking new plugin %p as registered",
@@ -844,8 +929,6 @@ gst_registry_scan_path_level (GstRegistry * registry, const gchar * path,
 
   g_dir_close (dir);
 
-  GST_DEBUG_OBJECT (registry, "registry changed? %d", changed);
-
   return changed;
 }
 
@@ -871,19 +954,22 @@ gst_registry_scan_path (GstRegistry * registry, const gchar * path)
   GST_DEBUG_OBJECT (registry, "scanning path %s", path);
   changed = gst_registry_scan_path_level (registry, path, 10);
 
-  GST_DEBUG_OBJECT (registry, "registry changed? %d", changed);
+  GST_DEBUG_OBJECT (registry, "registry changed in path %s: %d", path, changed);
 
   return changed;
 }
 
-void
-_gst_registry_remove_cache_plugins (GstRegistry * registry)
+/* Unref all plugins marked 'cached', to clear old plugins that no
+ * longer exist. Returns TRUE if any plugins were removed */
+gboolean
+_priv_gst_registry_remove_cache_plugins (GstRegistry * registry)
 {
   GList *g;
   GList *g_next;
   GstPlugin *plugin;
+  gboolean changed = FALSE;
 
-  g_return_if_fail (GST_IS_REGISTRY (registry));
+  g_return_val_if_fail (GST_IS_REGISTRY (registry), FALSE);
 
   GST_OBJECT_LOCK (registry);
 
@@ -895,16 +981,17 @@ _gst_registry_remove_cache_plugins (GstRegistry * registry)
     if (plugin->flags & GST_PLUGIN_FLAG_CACHED) {
       GST_DEBUG_OBJECT (registry, "removing cached plugin \"%s\"",
           GST_STR_NULL (plugin->filename));
-      /* seems it would be sufficient just to do a delete_link for o(1) deletion
-       * -- we have to traverse the whole list anyway, and dup entries (if
-       * possible) should have dup refcounts */
-      registry->plugins = g_list_remove (registry->plugins, plugin);
+      registry->plugins = g_list_delete_link (registry->plugins, g);
+      gst_registry_remove_features_for_plugin_unlocked (registry, plugin);
       gst_object_unref (plugin);
+      changed = TRUE;
     }
     g = g_next;
   }
 
   GST_OBJECT_UNLOCK (registry);
+
+  return changed;
 }
 
 
@@ -922,7 +1009,8 @@ _gst_plugin_feature_filter_plugin_name (GstPluginFeature * feature,
  *
  * Retrieves a #GList of features of the plugin with name @name.
  *
- * Returns: a #GList of #GstPluginFeature. gst_plugin_feature_list_free() after usage.
+ * Returns: a #GList of #GstPluginFeature. Use gst_plugin_feature_list_free()
+ * after usage.
  */
 GList *
 gst_registry_get_feature_list_by_plugin (GstRegistry * registry,
@@ -935,8 +1023,9 @@ gst_registry_get_feature_list_by_plugin (GstRegistry * registry,
       _gst_plugin_feature_filter_plugin_name, FALSE, (gpointer) name);
 }
 
+/* Unref and delete the default registry */
 void
-_gst_registry_cleanup ()
+_priv_gst_registry_cleanup ()
 {
   GstRegistry *registry;
 
