@@ -20,7 +20,12 @@
  * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
  * Boston, MA 02111-1307, USA.
  */
-
+/**
+ * SECTION:element-identity
+ *
+ * Dummy element that passes incomming data through unmodified. It has some
+ * useful diagnostic functions, such as offset and timestamp checking.
+ */
 
 #ifdef HAVE_CONFIG_H
 #  include "config.h"
@@ -45,13 +50,6 @@ static GstStaticPadTemplate srctemplate = GST_STATIC_PAD_TEMPLATE ("src",
 GST_DEBUG_CATEGORY_STATIC (gst_identity_debug);
 #define GST_CAT_DEFAULT gst_identity_debug
 
-static const GstElementDetails gst_identity_details =
-GST_ELEMENT_DETAILS ("Identity",
-    "Generic",
-    "Pass data without modification",
-    "Erik Walthinsen <omega@cse.ogi.edu>");
-
-
 /* Identity signals and args */
 enum
 {
@@ -70,6 +68,9 @@ enum
 #define DEFAULT_DUMP                    FALSE
 #define DEFAULT_SYNC                    FALSE
 #define DEFAULT_CHECK_PERFECT           FALSE
+#define DEFAULT_CHECK_IMPERFECT_TIMESTAMP FALSE
+#define DEFAULT_CHECK_IMPERFECT_OFFSET    FALSE
+#define DEFAULT_SIGNAL_HANDOFFS           TRUE
 
 enum
 {
@@ -83,7 +84,10 @@ enum
   PROP_LAST_MESSAGE,
   PROP_DUMP,
   PROP_SYNC,
-  PROP_CHECK_PERFECT
+  PROP_CHECK_PERFECT,
+  PROP_CHECK_IMPERFECT_TIMESTAMP,
+  PROP_CHECK_IMPERFECT_OFFSET,
+  PROP_SIGNAL_HANDOFFS
 };
 
 
@@ -102,6 +106,9 @@ static void gst_identity_get_property (GObject * object, guint prop_id,
 static gboolean gst_identity_event (GstBaseTransform * trans, GstEvent * event);
 static GstFlowReturn gst_identity_transform_ip (GstBaseTransform * trans,
     GstBuffer * buf);
+static GstFlowReturn gst_identity_prepare_output_buffer (GstBaseTransform
+    * trans, GstBuffer * in_buf, gint out_size, GstCaps * out_caps,
+    GstBuffer ** out_buf);
 static gboolean gst_identity_start (GstBaseTransform * trans);
 static gboolean gst_identity_stop (GstBaseTransform * trans);
 
@@ -112,11 +119,14 @@ gst_identity_base_init (gpointer g_class)
 {
   GstElementClass *gstelement_class = GST_ELEMENT_CLASS (g_class);
 
+  gst_element_class_set_details_simple (gstelement_class,
+      "Identity",
+      "Generic",
+      "Pass data without modification", "Erik Walthinsen <omega@cse.ogi.edu>");
   gst_element_class_add_pad_template (gstelement_class,
       gst_static_pad_template_get (&srctemplate));
   gst_element_class_add_pad_template (gstelement_class,
       gst_static_pad_template_get (&sinktemplate));
-  gst_element_class_set_details (gstelement_class, &gst_identity_details);
 }
 
 static void
@@ -176,41 +186,72 @@ gst_identity_class_init (GstIdentityClass * klass)
   g_object_class_install_property (gobject_class, PROP_SLEEP_TIME,
       g_param_spec_uint ("sleep-time", "Sleep time",
           "Microseconds to sleep between processing", 0, G_MAXUINT,
-          DEFAULT_SLEEP_TIME, G_PARAM_READWRITE));
+          DEFAULT_SLEEP_TIME, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
   g_object_class_install_property (gobject_class, PROP_ERROR_AFTER,
-      g_param_spec_int ("error_after", "Error After", "Error after N buffers",
-          G_MININT, G_MAXINT, DEFAULT_ERROR_AFTER, G_PARAM_READWRITE));
-  g_object_class_install_property (gobject_class,
-      PROP_DROP_PROBABILITY, g_param_spec_float ("drop_probability",
-          "Drop Probability",
-          "The Probability a buffer is dropped (not implemented)", 0.0, 1.0,
-          DEFAULT_DROP_PROBABILITY, G_PARAM_READWRITE));
+      g_param_spec_int ("error-after", "Error After", "Error after N buffers",
+          G_MININT, G_MAXINT, DEFAULT_ERROR_AFTER,
+          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+  g_object_class_install_property (gobject_class, PROP_DROP_PROBABILITY,
+      g_param_spec_float ("drop-probability", "Drop Probability",
+          "The Probability a buffer is dropped", 0.0, 1.0,
+          DEFAULT_DROP_PROBABILITY,
+          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
   g_object_class_install_property (gobject_class, PROP_DATARATE,
       g_param_spec_int ("datarate", "Datarate",
           "(Re)timestamps buffers with number of bytes per second (0 = inactive)",
-          0, G_MAXINT, DEFAULT_DATARATE, G_PARAM_READWRITE));
+          0, G_MAXINT, DEFAULT_DATARATE,
+          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
   g_object_class_install_property (gobject_class, PROP_SILENT,
       g_param_spec_boolean ("silent", "silent", "silent", DEFAULT_SILENT,
-          G_PARAM_READWRITE));
+          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
   g_object_class_install_property (gobject_class, PROP_SINGLE_SEGMENT,
       g_param_spec_boolean ("single-segment", "Single Segment",
           "Timestamp buffers and eat newsegments so as to appear as one segment",
-          DEFAULT_SINGLE_SEGMENT, G_PARAM_READWRITE));
+          DEFAULT_SINGLE_SEGMENT, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
   g_object_class_install_property (gobject_class, PROP_LAST_MESSAGE,
       g_param_spec_string ("last-message", "last-message", "last-message", NULL,
-          G_PARAM_READABLE));
+          G_PARAM_READABLE | G_PARAM_STATIC_STRINGS));
   g_object_class_install_property (gobject_class, PROP_DUMP,
       g_param_spec_boolean ("dump", "Dump", "Dump buffer contents to stdout",
-          DEFAULT_DUMP, G_PARAM_READWRITE));
+          DEFAULT_DUMP, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
   g_object_class_install_property (gobject_class, PROP_SYNC,
       g_param_spec_boolean ("sync", "Synchronize",
-          "Synchronize to pipeline clock", DEFAULT_SYNC, G_PARAM_READWRITE));
+          "Synchronize to pipeline clock", DEFAULT_SYNC,
+          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
   g_object_class_install_property (gobject_class, PROP_CHECK_PERFECT,
       g_param_spec_boolean ("check-perfect", "Check For Perfect Stream",
-          "Verify that the stream is time- and data-contiguous",
-          DEFAULT_CHECK_PERFECT, G_PARAM_READWRITE));
+          "Verify that the stream is time- and data-contiguous. "
+          "This only logs in the debug log.  This will be deprecated in favor "
+          "of the check-imperfect-timestamp/offset properties.",
+          DEFAULT_CHECK_PERFECT, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+  g_object_class_install_property (gobject_class,
+      PROP_CHECK_IMPERFECT_TIMESTAMP,
+      g_param_spec_boolean ("check-imperfect-timestamp",
+          "Check for discontiguous timestamps",
+          "Send element messages if timestamps and durations do not match up",
+          DEFAULT_CHECK_IMPERFECT_TIMESTAMP,
+          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+  g_object_class_install_property (gobject_class, PROP_CHECK_IMPERFECT_OFFSET,
+      g_param_spec_boolean ("check-imperfect-offset",
+          "Check for discontiguous offset",
+          "Send element messages if offset and offset_end do not match up",
+          DEFAULT_CHECK_IMPERFECT_OFFSET,
+          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
 
-   /**
+  /**
+   * GstIdentity:signal-handoffs
+   *
+   * If set to #TRUE, the identity will emit a handoff signal when handling a buffer.
+   * When set to #FALSE, no signal will be emited, which might improve performance.
+   *
+   * Since: 0.10.16
+   */
+  g_object_class_install_property (gobject_class, PROP_SIGNAL_HANDOFFS,
+      g_param_spec_boolean ("signal-handoffs",
+          "Signal handoffs", "Send a signal before pushing the buffer",
+          DEFAULT_SIGNAL_HANDOFFS, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+
+  /**
    * GstIdentity::handoff:
    * @identity: the identity instance
    * @buffer: the buffer that just has been received
@@ -228,6 +269,8 @@ gst_identity_class_init (GstIdentityClass * klass)
   gstbasetrans_class->event = GST_DEBUG_FUNCPTR (gst_identity_event);
   gstbasetrans_class->transform_ip =
       GST_DEBUG_FUNCPTR (gst_identity_transform_ip);
+  gstbasetrans_class->prepare_output_buffer =
+      GST_DEBUG_FUNCPTR (gst_identity_prepare_output_buffer);
   gstbasetrans_class->start = GST_DEBUG_FUNCPTR (gst_identity_start);
   gstbasetrans_class->stop = GST_DEBUG_FUNCPTR (gst_identity_stop);
 }
@@ -235,8 +278,6 @@ gst_identity_class_init (GstIdentityClass * klass)
 static void
 gst_identity_init (GstIdentity * identity, GstIdentityClass * g_class)
 {
-  gst_base_transform_set_passthrough (GST_BASE_TRANSFORM (identity), TRUE);
-
   identity->sleep_time = DEFAULT_SLEEP_TIME;
   identity->error_after = DEFAULT_ERROR_AFTER;
   identity->drop_probability = DEFAULT_DROP_PROBABILITY;
@@ -245,8 +286,11 @@ gst_identity_init (GstIdentity * identity, GstIdentityClass * g_class)
   identity->single_segment = DEFAULT_SINGLE_SEGMENT;
   identity->sync = DEFAULT_SYNC;
   identity->check_perfect = DEFAULT_CHECK_PERFECT;
+  identity->check_imperfect_timestamp = DEFAULT_CHECK_IMPERFECT_TIMESTAMP;
+  identity->check_imperfect_offset = DEFAULT_CHECK_IMPERFECT_OFFSET;
   identity->dump = DEFAULT_DUMP;
   identity->last_message = NULL;
+  identity->signal_handoffs = DEFAULT_SIGNAL_HANDOFFS;
 }
 
 static gboolean
@@ -296,7 +340,14 @@ gst_identity_event (GstBaseTransform * trans, GstEvent * event)
     }
   }
 
-  GST_BASE_TRANSFORM_CLASS (parent_class)->event (trans, event);
+  /* Reset previous timestamp, duration and offsets on NEWSEGMENT
+   * to prevent false warnings when checking for perfect streams */
+  if (GST_EVENT_TYPE (event) == GST_EVENT_NEWSEGMENT) {
+    identity->prev_timestamp = identity->prev_duration = GST_CLOCK_TIME_NONE;
+    identity->prev_offset = identity->prev_offset_end = GST_BUFFER_OFFSET_NONE;
+  }
+
+  ret = parent_class->event (trans, event);
 
   if (identity->single_segment
       && (GST_EVENT_TYPE (event) == GST_EVENT_NEWSEGMENT)) {
@@ -307,6 +358,28 @@ gst_identity_event (GstBaseTransform * trans, GstEvent * event)
   return ret;
 }
 
+static GstFlowReturn
+gst_identity_prepare_output_buffer (GstBaseTransform * trans,
+    GstBuffer * in_buf, gint out_size, GstCaps * out_caps, GstBuffer ** out_buf)
+{
+  GstIdentity *identity = GST_IDENTITY (trans);
+
+  /* only bother if we may have to alter metadata */
+  if (identity->datarate > 0 || identity->single_segment) {
+    if (gst_buffer_is_metadata_writable (in_buf))
+      *out_buf = gst_buffer_ref (in_buf);
+    else {
+      /* make even less writable */
+      gst_buffer_ref (in_buf);
+      /* extra ref is dropped going through the official process */
+      *out_buf = gst_buffer_make_metadata_writable (in_buf);
+    }
+  } else
+    *out_buf = gst_buffer_ref (in_buf);
+
+  return GST_FLOW_OK;
+}
+
 static void
 gst_identity_check_perfect (GstIdentity * identity, GstBuffer * buf)
 {
@@ -337,7 +410,9 @@ gst_identity_check_perfect (GstIdentity * identity, GstBuffer * buf)
       }
 
       offset = GST_BUFFER_OFFSET (buf);
-      if (identity->prev_offset_end != offset) {
+      if (identity->prev_offset_end != offset &&
+          identity->prev_offset_end != GST_BUFFER_OFFSET_NONE &&
+          offset != GST_BUFFER_OFFSET_NONE) {
         GST_WARNING_OBJECT (identity,
             "Buffer not data-contiguous with previous one: "
             "prev offset_end %" G_GINT64_FORMAT ", new offset %"
@@ -347,11 +422,103 @@ gst_identity_check_perfect (GstIdentity * identity, GstBuffer * buf)
       GST_DEBUG_OBJECT (identity, "can't check time-contiguity, no timestamp "
           "and/or duration were set on previous buffer");
     }
+  }
+}
 
-    /* update prev values */
-    identity->prev_timestamp = timestamp;
-    identity->prev_duration = GST_BUFFER_DURATION (buf);
-    identity->prev_offset_end = GST_BUFFER_OFFSET_END (buf);
+static void
+gst_identity_check_imperfect_timestamp (GstIdentity * identity, GstBuffer * buf)
+{
+  GstClockTime timestamp = GST_BUFFER_TIMESTAMP (buf);
+
+  /* invalid timestamp drops us out of check.  FIXME: maybe warn ? */
+  if (timestamp != GST_CLOCK_TIME_NONE) {
+    /* check if we had a previous buffer to compare to */
+    if (identity->prev_timestamp != GST_CLOCK_TIME_NONE &&
+        identity->prev_duration != GST_CLOCK_TIME_NONE) {
+      guint64 offset;
+      GstClockTime t_expected;
+      GstClockTimeDiff dt;
+
+      offset = GST_BUFFER_OFFSET (buf);
+      t_expected = identity->prev_timestamp + identity->prev_duration;
+      dt = GST_CLOCK_DIFF (t_expected, timestamp);
+      if (dt != 0) {
+        /*
+         * "imperfect-timestamp" bus message:
+         * @identity:        the identity instance
+         * @prev-timestamp:  the previous buffer timestamp
+         * @prev-duration:   the previous buffer duration
+         * @prev-offset:     the previous buffer offset
+         * @prev-offset-end: the previous buffer offset end
+         * @cur-timestamp:   the current buffer timestamp
+         * @cur-duration:    the current buffer duration
+         * @cur-offset:      the current buffer offset
+         * @cur-offset-end:  the current buffer offset end
+         *
+         * This bus message gets emitted if the check-imperfect-timestamp
+         * property is set and there is a gap in time between the
+         * last buffer and the newly received buffer.
+         */
+        gst_element_post_message (GST_ELEMENT (identity),
+            gst_message_new_element (GST_OBJECT (identity),
+                gst_structure_new ("imperfect-timestamp",
+                    "prev-timestamp", G_TYPE_UINT64,
+                    identity->prev_timestamp, "prev-duration", G_TYPE_UINT64,
+                    identity->prev_duration, "prev-offset", G_TYPE_UINT64,
+                    identity->prev_offset, "prev-offset-end", G_TYPE_UINT64,
+                    identity->prev_offset_end, "cur-timestamp", G_TYPE_UINT64,
+                    timestamp, "cur-duration", G_TYPE_UINT64,
+                    GST_BUFFER_DURATION (buf), "cur-offset", G_TYPE_UINT64,
+                    GST_BUFFER_OFFSET (buf), "cur-offset-end", G_TYPE_UINT64,
+                    GST_BUFFER_OFFSET_END (buf), NULL)));
+      }
+    } else {
+      GST_DEBUG_OBJECT (identity, "can't check data-contiguity, no "
+          "offset_end was set on previous buffer");
+    }
+  }
+}
+
+static void
+gst_identity_check_imperfect_offset (GstIdentity * identity, GstBuffer * buf)
+{
+  guint64 offset;
+
+  offset = GST_BUFFER_OFFSET (buf);
+
+  if (identity->prev_offset_end != offset &&
+      identity->prev_offset_end != GST_BUFFER_OFFSET_NONE &&
+      offset != GST_BUFFER_OFFSET_NONE) {
+    /*
+     * "imperfect-offset" bus message:
+     * @identity:        the identity instance
+     * @prev-timestamp:  the previous buffer timestamp
+     * @prev-duration:   the previous buffer duration
+     * @prev-offset:     the previous buffer offset
+     * @prev-offset-end: the previous buffer offset end
+     * @cur-timestamp:   the current buffer timestamp
+     * @cur-duration:    the current buffer duration
+     * @cur-offset:      the current buffer offset
+     * @cur-offset-end:  the current buffer offset end
+     *
+     * This bus message gets emitted if the check-imperfect-offset
+     * property is set and there is a gap in offsets between the
+     * last buffer and the newly received buffer.
+     */
+    gst_element_post_message (GST_ELEMENT (identity),
+        gst_message_new_element (GST_OBJECT (identity),
+            gst_structure_new ("imperfect-offset", "prev-timestamp",
+                G_TYPE_UINT64, identity->prev_timestamp, "prev-duration",
+                G_TYPE_UINT64, identity->prev_duration, "prev-offset",
+                G_TYPE_UINT64, identity->prev_offset, "prev-offset-end",
+                G_TYPE_UINT64, identity->prev_offset_end, "cur-timestamp",
+                G_TYPE_UINT64, GST_BUFFER_TIMESTAMP (buf), "cur-duration",
+                G_TYPE_UINT64, GST_BUFFER_DURATION (buf), "cur-offset",
+                G_TYPE_UINT64, GST_BUFFER_OFFSET (buf), "cur-offset-end",
+                G_TYPE_UINT64, GST_BUFFER_OFFSET_END (buf), NULL)));
+  } else {
+    GST_DEBUG_OBJECT (identity, "can't check offset contiguity, no offset "
+        "and/or offset_end were set on previous buffer");
   }
 }
 
@@ -364,6 +531,16 @@ gst_identity_transform_ip (GstBaseTransform * trans, GstBuffer * buf)
 
   if (identity->check_perfect)
     gst_identity_check_perfect (identity, buf);
+  if (identity->check_imperfect_timestamp)
+    gst_identity_check_imperfect_timestamp (identity, buf);
+  if (identity->check_imperfect_offset)
+    gst_identity_check_imperfect_offset (identity, buf);
+
+  /* update prev values */
+  identity->prev_timestamp = GST_BUFFER_TIMESTAMP (buf);
+  identity->prev_duration = GST_BUFFER_DURATION (buf);
+  identity->prev_offset_end = GST_BUFFER_OFFSET_END (buf);
+  identity->prev_offset = GST_BUFFER_OFFSET (buf);
 
   if (identity->error_after >= 0) {
     identity->error_after--;
@@ -391,10 +568,8 @@ gst_identity_transform_ip (GstBaseTransform * trans, GstBuffer * buf)
         GST_OBJECT_UNLOCK (identity);
         g_object_notify (G_OBJECT (identity), "last-message");
       }
-      /* FIXME, this does not drop the buffer in basetransform. Actually
-       * dropping the buffer in transform_ip is not possible without a new
-       * custom GstFlowReturn value. */
-      return GST_FLOW_OK;
+      /* return DROPPED to basetransform. */
+      return GST_BASE_TRANSFORM_FLOW_DROPPED;
     }
   }
 
@@ -419,15 +594,17 @@ gst_identity_transform_ip (GstBaseTransform * trans, GstBuffer * buf)
   }
 
   if (identity->datarate > 0) {
-    GstClockTime time = identity->offset * GST_SECOND / identity->datarate;
+    GstClockTime time = gst_util_uint64_scale_int (identity->offset,
+        GST_SECOND, identity->datarate);
 
     GST_BUFFER_TIMESTAMP (buf) = time;
     GST_BUFFER_DURATION (buf) =
         GST_BUFFER_SIZE (buf) * GST_SECOND / identity->datarate;
   }
 
-  g_signal_emit (G_OBJECT (identity), gst_identity_signals[SIGNAL_HANDOFF], 0,
-      buf);
+  if (identity->signal_handoffs)
+    g_signal_emit (G_OBJECT (identity), gst_identity_signals[SIGNAL_HANDOFF], 0,
+        buf);
 
   if (trans->segment.format == GST_FORMAT_TIME)
     runtimestamp = gst_segment_to_running_time (&trans->segment,
@@ -512,6 +689,15 @@ gst_identity_set_property (GObject * object, guint prop_id,
     case PROP_CHECK_PERFECT:
       identity->check_perfect = g_value_get_boolean (value);
       break;
+    case PROP_CHECK_IMPERFECT_TIMESTAMP:
+      identity->check_imperfect_timestamp = g_value_get_boolean (value);
+      break;
+    case PROP_CHECK_IMPERFECT_OFFSET:
+      identity->check_imperfect_offset = g_value_get_boolean (value);
+      break;
+    case PROP_SIGNAL_HANDOFFS:
+      identity->signal_handoffs = g_value_get_boolean (value);
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -559,6 +745,15 @@ gst_identity_get_property (GObject * object, guint prop_id, GValue * value,
     case PROP_CHECK_PERFECT:
       g_value_set_boolean (value, identity->check_perfect);
       break;
+    case PROP_CHECK_IMPERFECT_TIMESTAMP:
+      g_value_set_boolean (value, identity->check_imperfect_timestamp);
+      break;
+    case PROP_CHECK_IMPERFECT_OFFSET:
+      g_value_set_boolean (value, identity->check_imperfect_offset);
+      break;
+    case PROP_SIGNAL_HANDOFFS:
+      g_value_set_boolean (value, identity->signal_handoffs);
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -575,7 +770,8 @@ gst_identity_start (GstBaseTransform * trans)
   identity->offset = 0;
   identity->prev_timestamp = GST_CLOCK_TIME_NONE;
   identity->prev_duration = GST_CLOCK_TIME_NONE;
-  identity->prev_offset_end = -1;
+  identity->prev_offset_end = GST_BUFFER_OFFSET_NONE;
+  identity->prev_offset = GST_BUFFER_OFFSET_NONE;
 
   return TRUE;
 }
